#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for stories
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions
import pyscribus.marks as marks
import pyscribus.notes as notes

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class StoryEnding(PyScribusElement):
    """
    Ending marker in Scribus stories (trail)
    """

    def __init__(self, **kwargs):
        super().__init__()
        self.alignment = None
        self.parent = None

    def toxml(self):
        xml = ET.Element("trail")

        if self.alignment is not None:
            xml.attrib["ALIGN"] = self.alignment

        if self.parent is not None:
            xml.attrib["PARENT"] = self.parent

        return xml

    def fromxml(self, xml):
        if xml.tag == "trail":

            align = xml.get("ALIGN")
            if align is not None:
                # TODO use human values...
                self.alignment = align

            parent = xml.get("PARENT")
            if parent is not None:
                self.parent = parent

            return True

        return False


class StoryDefaultStyle(OrphanElement):
    """
    Default style marker in Scribus stories (DefaultStyle)
    """

    def __init__(self):
        OrphanElement.__init__(self, "DefaultStyle")


class StoryLineBreak(OrphanElement):
    """
    Line break marker in Scribus stories (breakline)
    """

    def __init__(self):
        OrphanElement.__init__(self, "breakline")


class NonBreakingHyphen(OrphanElement):
    """
    Non breaking hyphen in Scribus stories (nbhyphen)
    """

    def __init__(self):
        OrphanElement.__init__(self, "nbhyphen")


class NonBreakingSpace(OrphanElement):
    """
    Non breaking space in Scribus stories (nbspace)
    """

    def __init__(self):
        OrphanElement.__init__(self, "nbspace")


class StoryParagraphEnding(PyScribusElement):
    """
    End of paragraph marker in Scribus stories (para)

    :type parent: string
    :param parent: Name of the paragraph style
    :type doc_parent: pyscribus.document.Document
    :param parent: Parent Document instance

    :ivar string parent: Name of the paragraph style
    """

    def __init__(self, parent=False, doc_parent=False):
        super().__init__()

        self.parent = parent

        self.doc_parent = False

    def __repr__(self):
        return "PARAGEND|{}".format(self.parent)

    def fromxml(self, xml, check_style=True):
        parentpar = xml.get("PARENT")
        if parentpar is None:
            self.parent = False
        else:
            self.parent = parentpar

            if check_style and self.parent:

                if self.doc_parent:

                    checked = False

                    for para_style in self.doc_parent.styles["paragraph"]:
                        if para_style.name == self.parent:
                            checked = True
                            break

                    if not checked:
                        raise exceptions.UnknownStyleInStory(self.parent)

        return True

    def toxml(self):
        xml = ET.Element("para")

        if self.parent:
            xml.attrib["PARENT"] = self.parent

        return xml


class StoryFragment(PyScribusElement):
    """
    Text fragment (ITEXT) in Scribus stories.

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    :ivar string text: Text content
    :ivar dict font: Font details
    :ivar dict features: Font special formatting

    +----------------+---------------------------------+--------------+
    | Kwargs         | Setting                         | Type         |
    +================+=================================+==============+
    | text           | Fragment text                   | string       |
    +----------------+---------------------------------+--------------+
    | font           | Font name                       | string       |
    +----------------+---------------------------------+--------------+
    | fontsize       | Font size                       | float        |
    +----------------+---------------------------------+--------------+
    | fontcolor      | Font color name                 | string       |
    +----------------+---------------------------------+--------------+
    | fontopacity    | Font opacity                    | float        |
    |                |                                 | (percentage) |
    +----------------+---------------------------------+--------------+
    | default        | Equivalent to a fromdefault     | boolean or   |
    |                | call, value being True or the   | string       |
    |                | default name                    |              |
    +----------------+---------------------------------+--------------+
    | features       | Fragment features               | dict         |
    +----------------+---------------------------------+--------------+
    | inherit        | Font feature : inherit          | boolean      |
    +----------------+---------------------------------+--------------+
    | smallcaps      | Font feature : small caps       | boolean      |
    +----------------+---------------------------------+--------------+
    | allcaps        | Font feature : all caps         | boolean      |
    +----------------+---------------------------------+--------------+
    | superscript    | Font feature : superscript      | boolean      |
    +----------------+---------------------------------+--------------+
    | strike         | Font feature : striked          | boolean      |
    +----------------+---------------------------------+--------------+
    | subscript      | Font feature : subscript        | boolean      |
    +----------------+---------------------------------+--------------+
    | underline      | Font feature : underlined       | boolean      |
    +----------------+---------------------------------+--------------+
    | underlinewords | Font feature : underlined words | boolean      |
    +----------------+---------------------------------+--------------+

    :Example:

    .. code:: python

       frag = stories.StoryFragment(
           text="Lorem ipsum", fontsize=7, fontcolor="Grey"
       )

    """

    def __init__(self, **kwargs):
        super().__init__()

        self.text = ""

        # To not export to XML. This is defined by the following
        # StoryParagraphEnding in Story.sequence
        self.paragraph_style = False

        self.font = {
            "name": False,
            "size": False,
            "color": False,
            "opacity": False
        }

        # Features
        self.features = {
            "inherit": False,
            "smallcaps": False,
            "allcaps": False,
            "superscript": False,
            "strike": False,
            "subscript": False,
            "underline": False,
            "underlinewords": False,
        }

        if kwargs:
            self._quick_setup(kwargs)

    def _quick_setup(self, settings):
        """
        Method for defining story fragment settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                if setting_name == "text":
                    self.text = setting_value

                if setting_value == "features":
                    self.features = setting_value

                if setting_name == "fontsize":
                    self.font["size"] = setting_value

                if setting_name == "fontcolor":
                    self.font["color"] = setting_value

                if setting_name == "fontopacity":
                    self.font["opacity"] = setting_value

                if setting_name == "font":
                    self.font["name"] = setting_value

                if setting_name in self.features.keys():

                    if setting_value:
                        self.features[setting_name] = True
                    else:
                        self.features[setting_name] = False

    def fromdefault(self):
        self.text = ""

        self.paragraph_style = False

        self.font = {
            "name": False,
            "size": False,
            "color": False,
            "opacity": False
        }

        self.features = {
            "inherit": False,
            "smallcaps": False,
            "allcaps": False,
            "superscript": False,
            "strike": False,
            "subscript": False,
            "underline": False,
            "underlinewords": False,
        }

    def __iadd__(self, fragment):
        """
        += operator can be used to join another fragment text.

        The second term **inherits the features of the first**.
        """
        if isinstance(fragment, StoryFragment):
            self.text += fragment.text
            return self
        else:
            raise TypeError()

    def __repr__(self):
        def font_repr(f):
            if f:
                return f
            else:
                return ""

        def features_repr(f):
            return [k for k in f if f[k]]

        return "FRAGMENT|{}|{}|{}".format(
            font_repr(self.font["name"]),
            self.text,
            features_repr(self.features)
        )

    def __str__(self):
        return self.text

    def toxml(self):
        xml = ET.Element("ITEXT")

        have_features = len([f for f in self.features.values() if f])

        if have_features:
            features = []

            for f,v in self.features.items():
                if v:
                    features.append(f)

            features = " ".join(features)

            xml.attrib["FEATURES"] = features

        for case in zip(
                    ["name", "color", "opacity", "size"],
                    ["FONT", "FCOLOR", "FSHADE", "FONTSIZE"],
                    ):
            if self.font[case[0]]:
                xml.attrib[case[1]] = self.font[case[0]]

        xml.attrib["CH"] = self.text

        return xml

    def fromxml(self, xml):
        fragtext = xml.get("CH")
        if fragtext is not None:

            # NOTE Don't do any strip, rstrip to @CH, as it may
            # contains legitimate spaces
            self.text = fragtext

            features = xml.get("FEATURES")
            if features is not None:
                self.set_features(features)

            for case in zip(
                        ["name", "color", "opacity", "size"],
                        ["FONT", "FCOLOR", "FSHADE", "FONTSIZE"],
                        ):

                att = xml.get(case[1])
                if att is not None:
                    self.font[case[0]] = att

            # TODO Reste de l’implémentation, puis :

            return True
        else:
            return False

    def set_features(self, features):
        # TODO

        features = features.split()

        for feature in features:
            if feature in self.features.keys():
                self.features[feature] = True


class StoryVariable(PyScribusElement):
    """
    Variable in a Scribus story (var).

    :type name: string
    :param name: Name of the variable. Must be in StoryVariable.var_names.

    :ivar string name: Name of the variable
    """

    var_names = ["pgno", "pgco"]

    def __init__(self, name=""):
        super().__init__()

        if nam in StoryVariable.var_names or nam == "":
            self.name = name

    def toxml(self):
        """
        :rtype: lxml.etree._Element
        :returns: XML representation of story variable
        """

        element = ET.Element("var")

        element.attrib["name"] = self.name

        return element

    def fromxml(self, xml):
        """
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        if xml.tag == "var":

            nam = xml.get("name")
            if nam is None:
                return False
            else:
                if nam in StoryVariable.var_names:
                    self.name = nam

            return True

        else:
            return False


class PageNumberVariable(StoryVariable):
    """
    Page number variable in a Scribus story
    """

    def __init__(self):
        super().__init__("pgno")

class PageCountVariable(StoryVariable):
    """
    Count of total pages variable in a Scribus story
    """

    def __init__(self):
        super().__init__("pgco")

# Variables globales 2 ==================================================#

variable_classes = {
    "pgno": PageNumberVariable,
    "pgco": PageCountVariable,
}

# Parsing functions =====================================================#

def sequencefromhtml(
        html,
        font={
            "italic": "Arial Italic",
            "bold": "Arial Bold",
            "bold-italic": "Arial Bold Italic"},
        alternate_emphasis=False,
        sla_document=False):
    """
    :type html: string
    :param html: Paragraph text in PSM (PyScribus Story Markup)
    :type font: dict
    :param font: Fonts used for text emphasis
    :type alternate_emphasis: boolean
    :param alternate_emphasis: Nested emphasis returns to regular font.
    :type sla_document: pyscribus.document.Document
    :param sla_document: Instance of PyScribus Document where notes & marks will
        be stored.
    :rtype: list
    :returns: List of pyscribus.stories.Story sequence elements
    """

    def tag_to_data(fragment, tag):
        """
        Modify fragment font and features according to tag value.
        """

        if tag in ["em", "i"]:
            fragment.font = font["italic"]

        if tag in ["b", "strong"]:
            fragment.font = font["bold"]

        if tag in ["sup"]:
            fragment.features["superscript"] = True

        if tag in ["sub"]:
            fragment.features["subscript"] = True

        if tag in ["u"]:
            fragment.features["underline"] = True

        if tag in ["sc"]:
            fragment.features["smallcaps"] = True

        return fragment

    def attrib_to_data(fragment, element):
        """
        Read @style attribute of element and modify fragment features
        accordingly.
        """

        if element.tag == "span":

            if "style" in element.attrib:
                css = [
                    r.strip()
                    for r
                    in element.attrib["style"].split(";")
                ]

                for rule in css:

                    # Little cleaning -> CSS property name and value
                    rule = rule.replace(": ",":").split(":")
                    name,value = rule[0],":".join(rule[1:])

                    if name == "font-variant":

                        if value == "small-caps":
                            fragment.features["smallcaps"] = True

                    if name == "text-transform":

                        if value == "uppercase":
                            fragment.features["allcaps"] = True

                    if name == "text-decoration":

                        if value == "underline":
                            fragment.features["underline"] = True

        return fragment

    def parse_note(element, sequence, sla_document):
        """
        Parse note XML data wether it comes from a <note> or a
        <span class="note"> element.
        """
        nid = element.get("id")

        if nid is None:
            raise ValueError(
                "<note> without @id."
            )
        else:
            # Mark of the note in the story ----------------------

            story_note = marks.StoryNoteMark(label=nid)
            story_note.fromdefault()

            # Note content in DOCUMENT ---------------------------

            # TODO FIXME Now that the mark is in the story,
            # we must add the note content into the DOCUMENT
            # <Notes>

            if sla_document:
                document_note = notes.Note()
                document_note.fromdefault()
                document_note.parent_mark = nid

                # TODO FIXME Handle note style
                # TODO FIXME Handle note text

                sla_document.notes.append(document_note)

            # Mark of the note in the DOCUMENT marks -------------

            if sla_document:
                document_mark = marks.DocumentMark()
                document_mark.fromdefault("note")
                document_mark.label = nid
                sla_document.marks.append(document_mark)

            # ----------------------------------------------------

            sequence.append(story_note)

        return sequence, sla_document

    def parse_element(element, sequence, font, previous, alternate_emphasis, sla_document):
        global variable_classes

        if element.tag in ["em", "i", "span", "b", "strong", "sup", "sub", "u", "sc"]:
            is_fragment = True

        if element.tag == "note":
            is_fragment = False

            sequence, sla_document = parse_note(element, sequence, sla_document)

        if element.tag in ["pgno", "pgco"]:
            is_fragment = False

            variable = Story.variable_classes[element.tag]()
            sequence.append(variable)

        if element.tag == "br":
            is_fragment = False
            sequence.append(StoryLineBreak())

        if element.tag == "span":

            if "class" in element.attrib:
                element_classes = [
                    cn.strip()
                    for cn in element.attrib["class"].strip().split()
                ]

                # Get the count of incompatible classes in span @class attribute
                # So <span class="pgno note"></span> will not work.
                main_class = len(
                    [
                        c
                        for c in ["pgno", "pgco", "note"]
                        if c in element_classes
                    ]
                )

                if main_class <= 1:

                    for element_class in element_classes:
                        # NOTE This iteration to check multiple classes in @class
                        # attribute, if a precision of a class is needed.

                        if element_class in ["pgno", "pgco"]:
                            is_fragment = False

                            variable = Story.variable_classes[element_class]()
                            sequence.append(variable)

                            break

                        if element_class == "note":
                            is_fragment = False

                            sequence, sla_document = parse_note(
                                element, sequence, sla_document
                            )

                            break
                else:
                    raise ValueError(
                        "Incompatible classes in span element @class : '{}'".format(
                            ",".join(element_classes)
                        )
                    )

        if is_fragment:
            fragment = StoryFragment(text=element.text)

            read_tag = True

            if alternate_emphasis:
                # If <em>lorem <em>ipsum</em> et dolor</em> must
                # produce an "ipsum" in regular style surrounded
                # by "lorem", "et dolor" both in italic.
                # Same thing for bold.

                if previous.tag in ["b", "strong", "i", "em"]:
                    if element.tag in ["b", "strong", "i", "em"]:
                        read_tag = False

            if read_tag:
                fragment = tag_to_data(fragment, element.tag)

            fragment = attrib_to_data(fragment, element)

            sequence.append(fragment)

        if len(element):

            for sub_element in element:

                sequence = parse_element(
                    sub_element,
                    sequence,
                    font,
                    element,
                    alternate_emphasis,
                    sla_document
                )

        if element.tail is not None:
            tail_fragment = StoryFragment(text=element.tail)
            tail_fragment = tag_to_data(tail_fragment, previous.tag)
            tail_fragment = attrib_to_data(tail_fragment, previous)

            sequence.append(tail_fragment)

        return sequence

    global variable_classes

    # Prepare parsing -----------------------------------------------

    # Encapsulate html in a <p> element if it doesn't exists, to
    # make lxml parsing work.

    if not html.startswith("<p>"):
        html = "<p>{}".format(html)

    if not html.endswith("</p>"):
        html = "{}</p>".format(html)

    parsed = ET.fromstring(html)

    # Parsing--------------------------------------------------------

    sequence = []

    if parsed.text is not None:
        sequence.append(StoryFragment(text=parsed.text))

    if len(parsed):

        for element in parsed:

            sequence = parse_element(
                element,
                sequence,
                font,
                parsed,
                alternate_emphasis,
                sla_document
            )

    if parsed.tail is not None:
        sequence.append(StoryFragment(text=parsed.tail))

    # ---------------------------------------------------------------

    return sequence

# Story class ===========================================================#

class Story(PyScribusElement):
    """
    Story in SLA (StoryText). A text buffer.

    :type sla_parent: pyscribus.sla.SLA
    :param sla_parent: Parent SLA instance.
    :type doc_parent: pyscribus.document.Document
    :param doc_parent: Parent Document instance.
    :type pgo_parent: pyscribus.pageobjects.PageObject
    :param pgo_parent: Parent page object instance.

    :ivar pyscribus.sla.SLA sla_parent: Parent SLA instance.
    :ivar pyscribus.document.Document doc_parent: Parent Document instance.
    :ivar list sequence: List of instances of StoryFragment,
        StoryParagraphEnding, StoryLineBreak, StoryDefaultStyle,
        NonBreakingHyphen, StoryVariable
    """

    def __init__(self, sla_parent=False, doc_parent=False, pgo_parent=False):
        super().__init__()

        self.sequence = []

        self.sla_parent = sla_parent
        self.doc_parent = doc_parent
        self.pgo_parent = pgo_parent

    def _without_ending(self):
        temp = []

        if self.sequence:
            if isinstance(self.sequence[-1], StoryEnding):
                temp = self.sequence[:-1]

        return temp

    def append_paragraph(self, **kwargs):
        """
        Append a paragraph at the end of the story.

        :type kwargs: dict
        :param kwargs: kwargs

        +-------------------+--------------+---------+--------------------------+
        | Kwargs key        | Kwargs value | Default |Usage                     |
        +===================+==============+=========+==========================+
        |text               | string       | False   | Paragraph text in PSM    |
        |                   |              |         | (PyScribus Story Markup) |
        +-------------------+--------------+---------+--------------------------+
        |ending             | boolean      | True    | Add a paragraph ending   |
        +-------------------+--------------+---------+--------------------------+
        |inherit_style      | boolean      | False   | Inherit paragraph style  |
        |                   |              |         | from last paragraph      |
        +-------------------+--------------+---------+--------------------------+
        |style              | string       | False   | Style of the paragraph.  |
        |                   |              |         | Overrides inherit_style. |
        +-------------------+--------------+---------+--------------------------+
        |alternate_emphasis | boolean      | False   | Nice emphasis, as nested |
        |                   |              |         | italic in italic becames |
        |                   |              |         | regular. Same for bold   |
        |                   |              |         | nested in bold.          |
        +-------------------+--------------+---------+--------------------------+

        As there is only one paragraph added by append_paragraph, the <p> element,
        enclosing the paragraph, can be omitted in text argument.

        :Example:

        .. code:: python

           story.append_paragraph(
               text="Le chant du cygne est très beau.", style="Normal"
           )
           story.append_paragraph(
               text="- Où as-tu <em>vu</em> ça ?", inherit_style=True
           )
           story.append_paragraph(
               text="<p>Et c'est ainsi qu'ils se quittèrent</p>",
               ending=False
           )

        """

        self._append_paragraph(kwargs)

        return True

    def append_paragraphs(self, paragraphs=[]):
        """
        Append multiple paragraphs at the end of the story.

        :type paragraphs: list
        :param paragraphs: List of Story.append_paragraph() kwargs dictionnaries.

        :Example:

        .. code:: python

           story.append_paragraphs(
              [
                  {"text": "A title", "style": "Title1"},
                  {"text": "Foreword", "style": "Foreword"},
                  {
                      "text": "First paragraph of content",
                      "style": "Normal"
                  },
                  {
                      "text": "Second paragraph of content",
                      "inherit_style": True
                  },
                  {"text": "03/01/2020", "ending": False}
              ]
           )

        .. seealso:: pyscribus.stories.Story.append_paragraph
        """

        if paragraphs:

            for paragraph in paragraphs:
                self._append_paragraph(paragraph)

            return True
        else:
            return False

    def _append_paragraph(self, kwargs):
        """
        Private method to append paragraph.

        Used by append_paragraph() and append_paragraphs()

        :type kwargs: dict
        :param paragraphs: kwargs of Story.append_paragraph()
        """

        # kwargs processing -------------------------------

        style,source,inherit_style = False,False,False
        alternate_emphasis = False

        ending = True

        font={
            "bold": "Arial Bold",
            "italic": "Arial Italic",
            "bold-italic": "Arial Bold Italic"
        }

        for param,value in kwargs.items():

            if param == "text":
                source = value

            if param == "ending":
                ending = bool(value)

            if param == "inherit_style":
                inherit_style = bool(value)

            if param == "style":
                style = value

            if param == "alternate_emphasis":
                alternate_emphasis = bool(value)

            if param == "font":

                if isinstance(value, dict):

                    for font_style in ["bold", "italic", "bold-italic"]:

                        if font_style not in value:

                            raise ValueError(
                                "Missing {} setting to font param".format(
                                    font_style
                                )
                            )

                    font = value

        # -------------------------------------------------

        sequence = sequencefromhtml(
            source,
            font=font,
            alternate_emphasis=alternate_emphasis,
            sla_document=self.doc_parent
        )

        if ending:
            para = StoryParagraphEnding()

            if style or inherit_style:
                ps = False

                # NOTE style overrides the inherited style

                if inherit_style:
                    # Get the last paragraph ending style

                    ps = [
                        e.parent
                        for e in self.sequence
                        if isinstance(e, StoryParagraphEnding)
                    ][-1]

                if style:
                    ps = style

                para.parent = ps

            sequence.append(para)

        temp = self._without_ending()
        self.sequence = temp + sequence
        self.end_contents()

    def toxml(self):
        """
        Return the story as XML element.

        :rtype: lxml._Element
        """

        xml = ET.Element("StoryText")

        for element in self.sequence:
            ex = element.toxml()
            xml.append(ex)

        return xml

    def fromxml(self, xml, check_style=True):
        """
        Parses XML of a SLA Story.

        :type xml: lxml._Element
        :param xml: SLA Story as lxml._Element
        :type check_style: bool
        :param check_style: Check if story paragraphs use known paragraph
            styles of Story.doc_parent Document. True by default.

        .. note:: You might set check_style to False if you want to parse
            stories independantly of any given SLA & Document instances.

        :rtype: bool
        :returns: bool

        """

        global variable_classes

        if xml.tag == "StoryText":

            for element in xml:

                if element.tag == "DefaultStyle":
                    self.sequence.append(StoryDefaultStyle())

                if element.tag == "ITEXT":

                    frag = StoryFragment()
                    success = frag.fromxml(element)

                    if success:
                        self.sequence.append(frag)

                # End of a paragraph

                if element.tag == "para":
                    para = StoryParagraphEnding(self.doc_parent)
                    success = para.fromxml(element, check_style)

                    if success:

                        # NOTE @PARENT carries the style name of the preceding
                        # StoryFragment / <ITEXT> in the sequence

                        if para.parent:

                            # NOTE Empty lines in the story are just <para>
                            # following each others. So we only set the
                            # paragraph_style of the preceding fragment…
                            # if there *is* a fragment to begin with.

                            if isinstance(self.sequence[-1], StoryFragment):
                                self.sequence[-1].paragraph_style = para.parent

                        self.sequence.append(para)

                # Line break in a paragraph

                if element.tag == "breakline":
                    self.sequence.append(StoryLineBreak())

                # Non breaking hyphen

                if element.tag == "nbhyphen":
                    self.sequence.append(NonBreakingHyphen())

                # Non breaking space

                if element.tag == "nbspace":
                    self.sequence.append(NonBreakingSpace())

                # Mark

                if element.tag == "MARK":
                    # TODO Selon @type, un StoryNoteMark ou autre
                    pass

                # Variable

                if element.tag == "var":

                    var_name = element.get("name")

                    if var_name is not None:
                        var_instance = Story.variable_classes[var_name]()
                        self.sequence.append(var_instance)

                # Story end

                if element.tag == "trail":
                    story_ending = StoryEnding()
                    success = story_ending.fromxml(element)

                    if success:
                        self.sequence.append(story_ending)
                    break

            return True

        return False

    def fromdefault(self):
        """
        Set a default story
        """

        self.sequence = [StoryDefaultStyle(), StoryEnding()]

    def append(self, element):
        """
        Append a element to the Story sequence.

        .. warning:: Don't use append() to append a StoryEnding, use
            end_contents() instead.

        :type element: StoryFragment, StoryParagraphEnding, StoryLineBreak, StoryDefaultStyle
        :param element: Element to add to the story sequence
        :rtype: bool
        :returns: bool

        .. seealso:: pyscribus.stories.end_contents
        """

        temp = self._without_ending()
        self.sequence = temp + [element]
        self.end_contents()

        return True

    def end_contents(self, no_trailing_paragraph=False):
        """
        Ends the story's content by making sure that the story sequence
        is valid.

        .. note:: It is better to use this method than appending a
            :class:`StoryEnding` instance directly into the story sequence.

        :type no_trailing_paragraph: bool
        :param no_trailing_paragraph: Remove last paragraph ending
        """
        story_ending = None

        if isinstance(self.sequence[-1], StoryEnding):
            story_ending = self.sequence.pop(-1)

        if no_trailing_paragraph:
            if isinstance(self.sequence[-1], StoryParagraphEnding):
                self.sequence.pop(-1)

        if story_ending is None:
            self.sequence.append(StoryEnding())
        else:
            self.sequence.append(story_ending)

    def init_contents(self):
        self.sequence = [StoryDefaultStyle()]

    def bypars(self):
        """
        Return a list of list of Story.sequence elements organized as human
        paragraphs.

        :returns: list of lists
        :rtype: ilst
        """

        pars = [] # List of paragraphs
        tmp = [] # Temporary list of paragraph elements

        for element in self.sequence:

            # End of the story -> last paragraph is added to the list

            if isinstance(element, StoryEnding):
                if tmp:
                    pars.append(tmp)
                    break

            # End of a paragraph ------------------------------------

            if isinstance(element, StoryParagraphEnding):

                if tmp:
                    tmp.append(element)
                    pars.append(tmp)
                    tmp = []

            # Paragraph content -------------------------------------

            for class_test in [StoryFragment, StoryDefaultStyle,
                    NonBreakingSpace, NonBreakingHyphen, StoryVariable]:

                if isinstance(element, class_test):
                    tmp.append(element)
                    break

        return pars

    def rawtext(self):
        """
        Returns a text string equivalent to *what Scribus story editor
        saves* as txt file.

        :rtype: string
        """
        text = ""

        for par in self.bypars():

            for element in par:

                # NOTE Text formatting is not saved
                if isinstance(element, StoryFragment):
                    text += element.text

                if isinstance(element, NonBreakingSpace):
                    text += " "

                # NOTE Line breaks are exported by Scribus... as spaces
                if isinstance(element, StoryLineBreak):
                    text += " "

                if isinstance(element, StoryParagraphEnding):
                    text += "\n"

        return text

    def templatable(self):
        """
        Return elements of Story sequence that are available for templating.

        :rtype: list
        :returns: List of pyscribus.stories.StoryFragment
        """

        contents = []

        pattern = self.sla_parent.templating["intext-pattern"]

        for element in self.sequence:

            if isinstance(element, StoryFragment):

                if pattern.findall(element.text):
                    contents.append(element)

        return contents

    def feed_templatable(self, datas={}):
        """
        """

        elements = self.templatable()

        if elements:
            modified = []

            for element in elements:

                for template_key, template_value in datas.items():

                    if template_key not in element.text:
                        continue

                    element.text = element.text.replace(
                        template_key, template_value
                    )

                    modified.append(element)

            return modified

        return False

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Classes for SLA Pages / Master pages.
"""

# Scribus pages are  arranged in a vast scratch space,  where 1 unit ==
# 1/72 inch.  Positive x is to  the right, and positive  y is downward.
# The positions, width and height in  the elements are in scratch space
# coordinates  with the  exception  of POCOOR  and COCOOR  coordinates,
# which are in the rotated and translated space of a PAGEOBJECT.

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.logs as logs
import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class PageAbstract(PyScribusElement):
    """
    Abstract class for pages and master pages.

    Use Page or MasterPage classes instead.

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table in Page, MasterPage)

    .. seealso:: :class:`MasterPage`, :class:`Page`
    """

    orientation_xml = {"portrait": "0", "landscape": "1"}
    autoguides_origin_xml = {"page": "0", "margins": "1"}

    effect_source_xml = {"internal": "0", "external": "1"}
    effect_mobile_line_xml = {"horizontal": "0", "vertical": 1}

    effect_type_xml = {
        "none": "0",
        "masks": "1",
        "box": "2",
        "dissolve": "3",
        "glitter": "4",
        "break": "5",
        "delete": "6"
    }

    # NOTE These directions are in the correct order, but Scribus only allows
    # the user to select "left-to-right" in its GUI.
    effect_direction_xml = {
        "left-to-right": "0",
        "top-to-bottom": "1",
        "bottom-to-top": "2",
        "right-to-left": "3",
        "tl-to-br": "4",
    }

    def __init__(self, **kwargs):
        PyScribusElement.__init__(self)

        self.box = dimensions.DimBox()

        self.borders = {
            "top": dimensions.Dim(40),
            "left": dimensions.Dim(40),
            "right": dimensions.Dim(40),
            "bottom": dimensions.Dim(40),
        }

        self.name = ""
        self.master_name = ""

        self.orientation = "portrait"

        self.paper_size = ""

        self.number = -1

        #--- FIXME Tester pour savoir comment constituer les listes ------

        self.guides = {
            "horizontal": [],
            "vertical": []
        }

        #--- FIXME Not documented ----------------------------------------

        # For multipage spreads, which page in the spread is the left most
        # LEFT="0"

        # PRESET="0"

        #--- Auto guides -------------------------------------------------

        # FIXME Understand AGSelection usage
        # AGSelection="0 0 0 0"

        self.auto_guides = {
            "lines": {
                "count": 0,
                "gap": dimensions.Dim(0),
                "origin": "page"
            },
            "columns": {
                "count": 0,
                "gap": dimensions.Dim(0),
                "origin": "page"
            },
            "selection": []
        }

        #--- PDF effects -------------------------------------------------

        self.effect = {
            # @pageEffectDuration
            "duration": dimensions.Dim(1, "sec"),
            # @pageViewDuration
            "view-duration": dimensions.Dim(1, "sec"),
            # @effectType
            "type": "none",
            # @Dm
            "mobile-lines": "horizontal",
            # @M
            "source": "internal",
            # @Di
            "direction": "left-to-right"
        }

        #--- FIXME Not documented ----------------------------------------

        # Same as DOCUMENT/PDF/Effekte (PDF display effects) attributes ?

        #-----------------------------------------------------------------

        if kwargs:
            self._quick_setup(kwargs)

    def _quick_setup(self, settings):
        """
        Method for defining (master)page settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                # (Master) page box -------------------------------------------

                if setting_name == "posx":
                    # self.box.coords["top-left"][0].value = float(setting_value)
                    self.box.setx("top-left", float(setting_value))

                if setting_name == "posy":
                    # self.box.coords["top-left"][1].value = float(setting_value)
                    self.box.sety("top-left", float(setting_value))

                if setting_name == "width":
                    self.box.dims["width"].value = float(setting_value)

                if setting_name == "height":
                    self.box.dims["height"].value = float(setting_value)

                # Borders -----------------------------------------------------

                if setting_name in [
                        "rightborder", "leftborder",
                        "topborder", "bottomborder"]:
                    side = setting_name.split("border")[0]
                    self.borders[side].value = float(setting_value)

                if setting_name == "borders":
                    # NOTE We solve borders settings like CSS margin property
                    # If only one value : same border for all sides
                    # If two values: vertical and horizontal borders
                    # If three values: top horizontal bottom
                    # If four values: top right bottom left

                    if isinstance(setting_value, list):
                        setting_len = len(setting_value)

                        if setting_len == 1:

                            for side in ["top", "right", "bottom", "left"]:
                                self.borders[side].value = float(
                                    setting_value[0]
                                )

                        if setting_len == 2:
                            sides = zip(
                                [["top", "bottom"], ["right", "left"]],
                                setting_value
                            )

                            for side in sides:
                                for s in side[0]:
                                    self.borders[s].value = float(side[1])

                        if setting_len == 3:
                            self.borders["top"].value = float(setting_value[0])

                            for side in ["right", "left"]:
                                self.borders[side].value = float(
                                    setting_value[1]
                                )

                            self.borders["bottom"].value = float(
                                setting_value[2]
                            )

                        if setting_len == 4:
                            sides = zip(
                                ["top", "right", "bottom", "left"],
                                setting_value
                            )

                            for side in sides:
                                self.borders[side[0]].value = float(side[1])

                # -------------------------------------------------------------

    def set_orientation(self, orientation):
        """
        Set (master) page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: string
        :returns: boolean
        :rtype: boolean
        """

        if orientation.lower() in PageAbstract.orientation_xml.keys():
            self.orientation = orientation.lower()
            # FIXME TODO Modify (master) page box accordingly

            return True

        else:
            raise ValueError(
                "orientation parameter of set_orientation must be "\
                "portrait' or 'landscape'"
            )

    def fromxml(self, xml, master=False):
        """
        Set (master) page attributes according to LXML Element

        :param xml: (Master) page source as XML element
        :type xml: lxml._Element
        :param master: If the page is a master page or not
        :type master: boolean
        :returns: boolean
        :rtype: boolean
        """
        if master:
            tag = "MASTERPAGE"
        else:
            tag = "PAGE"

        if xml.tag == tag:
            #--- Box settings -------------------------------------

            posx = xml.get("PAGEXPOS")
            posy = xml.get("PAGEYPOS")
            dim_width = xml.get("PAGEWIDTH")
            dim_height = xml.get("PAGEHEIGHT")

            valid_box = 0

            for test in [posx, posy, dim_width, dim_height]:
                if test is not None:
                    valid_box += 1

            if valid_box == 4:

                self.box.set_box(
                    top_lx=posx, top_ly=posy,
                    width=dim_width, height=dim_height
                )

            #--- Name / Master page name --------------------------

            self.name = xml.get("NAM")
            self.master_name = xml.get("MNAM")

            #--- Page number --------------------------------------

            page_number = xml.get("NUM")

            if page_number is not None:
                # NOTE Page 0 is the first page so we adjust to a human
                # readable count
                self.number = int(page_number) + 1

            #--- Paper size name and orientation ------------------

            paper_size = xml.get("Size")
            if paper_size is not None:
                self.paper_size = paper_size

            orientation = xml.get("Orientation")
            if orientation is not None:

                for h,x in PageAbstract.orientation_xml.items():
                    if orientation == x:
                        self.orientation = h
                        break

            #--- Borders ------------------------------------------

            for b in ["left", "top", "bottom", "right"]:
                att = "BORDER{}".format(b.upper())
                self.borders[b] = dimensions.Dim(float(xml.get(att)))

            #--- Guides -------------------------------------------

            for guide_type in ["vertical", "horizontal"]:

                att_name = "{}Guides".format(guide_type.capitalize())

                att = xml.get(att_name)
                if att is not None:
                    if att:
                        # NOTE Example of att value : "42.5197 56.6929 "

                        guides_list = []
                        guides_values = [g.strip() for g in att.split(" ")]

                        for g in guides_values:

                            if g:

                                guides_list.append(
                                    dimensions.Dim(float(g))
                                )

                        # self.guides[guide_type] = att
                        self.guides[guide_type] = guides_list

            #--- Auto guides --------------------------------------

            for case in [["lines", "horizontal"], ["columns", "vertical"]]:
                ag_count = xml.get("AG{}AutoCount".format(case[1]))
                ag_gap = xml.get("AG{}AutoGap".format(case[1]))
                ag_origin = xml.get("AG{}AutoRefer".format(case[1]))

                if ag_count is not None:
                    ag_count = int(ag_count)
                    self.auto_guides[case[0]]["count"] = ag_count

                if ag_gap is not None:
                    ag_gap = float(ag_gap)
                    self.auto_guides[case[0]]["gap"].value = ag_gap

                if ag_origin is not None:

                    for human,code in PageAbstract.autoguides_origin_xml.items():
                        if ag_origin == code:
                            self.auto_guides[case[0]]["origin"] = human

            #--- PDF effects -------------------------------------------------

            duration = xml.get("pageEffectDuration")
            if duration is not None:
                self.effect["duration"].value = int(duration)

            view_duration = xml.get("pageViewDuration")
            if view_duration is not None:
                self.effect["view-duration"].value = int(view_duration)

            effect_type = xml.get("effectType")
            if effect_type is not None:
                for human, code in PageAbstract.effect_type_xml.items():
                    if effect_type == code:
                        self.effect["type"] = human
                        break

            effect_lines = xml.get("Dm")
            if effect_lines is not None:
                for human, code in PageAbstract.effect_mobile_line_xml.items():
                    if effect_lines == code:
                        self.effect["mobile-lines"] = human
                        break

            effect_source = xml.get("M")
            if effect_source is not None:
                for human, code in PageAbstract.effect_source_xml.items():
                    if effect_source == code:
                        self.effect["source"] = human
                        break

            effect_direction = xml.get("Di")
            if effect_direction is not None:
                for human, code in PageAbstract.effect_direction_xml.items():
                    if effect_direction == code:
                        self.effect["direction"] = human
                        break

            #--- FIXME This records undocumented attributes -------

            self.undocumented = all_undocumented_to_python(xml)
            # self.undocumented = undocumented_to_python(
                # xml,
                # [
                    # "LEFT", "PRESET",
                    # "AGSelection",
                    # "pageEffectDuration", "pageViewDuration", "effectType", # "Dm", "M", "Di"
                # ]
            # )

            return True

        return False

    def toxml(self, master=False):
        """
        Returns (master) page as LXML Element

        :param master: If the page is a master page or not
        :type master: bool
        :returns: xml
        :rtype: lxml._Element
        """

        #--- Appropriate tag if master page or page -----------

        if master:
            tag = "MASTERPAGE"
        else:
            tag = "PAGE"

        xml = ET.Element(tag)

        #--- Size and dimensions ------------------------------

        xml.attrib["PAGEXPOS"] = self.box.coords["top-left"][0].toxmlstr()
        xml.attrib["PAGEYPOS"] = self.box.coords["top-left"][1].toxmlstr()

        xml.attrib["PAGEWIDTH"] = self.box.dims["width"].toxmlstr()
        xml.attrib["PAGEHEIGHT"] = self.box.dims["height"].toxmlstr()

        #--- Borders ------------------------------------------

        for b in ["left", "right", "top", "bottom"]:
            att = "BORDER{}".format(b.upper())
            xml.attrib[att] = self.borders[b].toxmlstr(True)

         #--- Page number --------------------------------------

        if self.number > 0:
            # NOTE Page 0 is the first page so we adjust back
            # from human counting to computer counting
            xml.attrib["NUM"] = str(self.number - 1)

        #--- Page name and master page name -------------------

        xml.attrib["NAM"] = self.name
        xml.attrib["MNAM"] = self.master_name

        #--- (Master) page size name and orientation ----------

        xml.attrib["Size"] = self.paper_size
        xml.attrib["Orientation"] = PageAbstract.orientation_xml[self.orientation]

        #--- Guides -------------------------------------------

        for guide_type,guides in self.guides.items():
            att_name = "{}Guides".format(guide_type.capitalize())

            # NOTE Example of att value : "42.5197 56.6929 "

            guides_str = " ".join([g.toxmlstr(True) for g in guides])

            if guides_str.strip():
                xml.attrib[att_name] = guides_str
            else:
                xml.attrib[att_name] = ""

        #--- Auto guides --------------------------------------

        for case in [["lines", "horizontal"], ["columns", "vertical"]]:
            # Page lines / columns count
            # AGhorizontalAutoCount="2"
            # AGverticalAutoCount="3"
            count_att = "AG{}AutoCount".format(case[1])

            # Page lines / columns gap (gouttière)
            # AGhorizontalAutoGap="17.007874015748"
            # AGverticalAutoGap="22.6771653543307"
            gap_att = "AG{}AutoGap".format(case[1])

            # Page lines / columns origin
            # 0 = Page 1 = Margins
            # AGhorizontalAutoRefer="0"
            # AGverticalAutoRefer="0"
            orig_att = "AG{}AutoRefer".format(case[1])

            xml.attrib[count_att] = str(self.auto_guides[case[0]]["count"])
            xml.attrib[gap_att] = self.auto_guides[case[0]]["gap"].toxmlstr()
            xml.attrib[orig_att] = PageAbstract.autoguides_origin_xml[self.auto_guides[case[0]]["origin"]]

        #--- PDF effects -------------------------------------------------

        xml.attrib["pageEffectDuration"] = self.effect["duration"].toxmlstr()
        xml.attrib["pageViewDuration"] = self.effect["view-duration"].toxmlstr()
        xml.attrib["effectType"] = PageAbstract.effect_type_xml[self.effect["type"]]
        xml.attrib["Dm"] = PageAbstract.effect_mobile_line_xml[self.effect["mobile-lines"]]
        xml.attrib["M"] = PageAbstract.effect_source_xml[self.effect["source"]]
        xml.attrib["Di"] = PageAbstract.effect_direction_xml[self.effect["direction"]]

        #--- FIXME This exports undocumented attributes -------

        try:
            # xml = undocumented_to_xml(xml, self.undocumented)
            xml, undoc_attribs = all_undocumented_to_xml(
                xml, self.undocumented, True, tag,
                logger=logs.getLogger()
            )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromdefault(self, master=False, default=False):
        """
        Set (master) page attributes according to known defaults.

        :param master: If the page is a master page or not
        :param default: Name of the set of defaults ("a4", "letter")
        :type master: bool
        :type default: str
        """

        for border in self.borders.values():
            border.value = 40

        dim_default = "a4"

        if default:
            for format_name in ["a4", "letter"]:
                if default.lower() == format_name:
                    dim_default = format_name
                    break

        self.master_name = "Normal"

        self.box.dims["width"].fromdefault("{}-width".format(dim_default))
        self.box.dims["height"].fromdefault("{}-height".format(dim_default))

        self.box.set_box(
            top_lx=100.0, top_ly=20.0,
            width=self.box.dims["width"].value,
            height=self.box.dims["height"].value
        )


class Page(PageAbstract):
    """
    Page in SLA

    To add objects to this page, use Document.append()

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+---------------------------------+------------+
    | Kwargs       | Setting                         | Value type |
    +==============+=================================+============+
    | default      | Equivalent to a ``fromdefault`` | boolean    |
    |              | call.                           | or string  |
    |              |                                 |            |
    |              | Value being True or the name of |            |
    |              | the default set.                |            |
    +--------------+---------------------------------+------------+
    | posx         | X position of the page          | float      |
    +--------------+---------------------------------+------------+
    | posy         | Y position of the page          | float      |
    +--------------+---------------------------------+------------+
    | width        | Page width                      | float      |
    +--------------+---------------------------------+------------+
    | height       | Page height                     | float      |
    +--------------+---------------------------------+------------+
    | borders      | Shorthand for :                 | List of    |
    |              |                                 | floats     |
    |              | *rightborder*, *leftborder*,    |            |
    |              | *topborder*, *bottomborder*     |            |
    |              |                                 |            |
    |              | Read like the CSS margin        |            |
    |              | property:                       |            |
    |              |                                 |            |
    |              | **With 1 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & right & bottom & left]   |            |
    |              |                                 |            |
    |              | **With 2 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & bottom, right & left]    |            |
    |              |                                 |            |
    |              | **With 3 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right & left, bottom]     |            |
    |              |                                 |            |
    |              | **With 4 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right, bottom, left]      |            |
    +--------------+---------------------------------+------------+
    | rightborder  | Right border                    | float      |
    +--------------+---------------------------------+------------+
    | leftborder   | Left border                     | float      |
    +--------------+---------------------------------+------------+
    | topborder    | Top border                      | float      |
    +--------------+---------------------------------+------------+
    | bottomborder | Bottom border                   | float      |
    +--------------+---------------------------------+------------+
    """

    def __init__(self, **kwargs):
        PageAbstract.__init__(self)
        PageAbstract._quick_setup(self, kwargs)

    def toxml(self):
        """
        Returns page as LXML Element

        :returns: xml
        :rtype: lxml._Element
        """
        return PageAbstract.toxml(self, False)

    def fromxml(self, xml):
        """
        Set page attributes according to LXML Element

        :param xml: Page source as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.fromxml(self, xml, False)

    def fromdefault(self, default=False):
        """
        Set page attributes according to known defaults.

        :param default: Name of the set of defaults ("a4", "letter")
        :type default: str
        """
        PageAbstract.fromdefault(self, False, default)

    def set_orientation(self, orientation):
        """
        Set page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: str
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.set_orientation(self, orientation)


class MasterPage(PageAbstract):
    """
    Master page in SLA

    To add objects to this master page, use Document.append()

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+---------------------------------+------------+
    | Kwargs       | Setting                         | Value type |
    +==============+=================================+============+
    | default      | Equivalent to a ``fromdefault`` | boolean    |
    |              | call.                           | or string  |
    |              |                                 |            |
    |              | Value being True or the name of |            |
    |              | the default set.                |            |
    +--------------+---------------------------------+------------+
    | posx         | X position of the page          | float      |
    +--------------+---------------------------------+------------+
    | posy         | Y position of the page          | float      |
    +--------------+---------------------------------+------------+
    | width        | Page width                      | float      |
    +--------------+---------------------------------+------------+
    | height       | Page height                     | float      |
    +--------------+---------------------------------+------------+
    | borders      | Shorthand for :                 | List of    |
    |              |                                 | floats     |
    |              | *rightborder*, *leftborder*,    |            |
    |              | *topborder*, *bottomborder*     |            |
    |              |                                 |            |
    |              | Read like the CSS margin        |            |
    |              | property:                       |            |
    |              |                                 |            |
    |              | **With 1 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & right & bottom & left]   |            |
    |              |                                 |            |
    |              | **With 2 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & bottom, right & left]    |            |
    |              |                                 |            |
    |              | **With 3 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right & left, bottom]     |            |
    |              |                                 |            |
    |              | **With 4 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right, bottom, left]      |            |
    +--------------+---------------------------------+------------+
    | rightborder  | Right border                    | float      |
    +--------------+---------------------------------+------------+
    | leftborder   | Left border                     | float      |
    +--------------+---------------------------------+------------+
    | topborder    | Top border                      | float      |
    +--------------+---------------------------------+------------+
    | bottomborder | Bottom border                   | float      |
    +--------------+---------------------------------+------------+
    """

    def __init__(self, **kwargs):
        PageAbstract.__init__(self)
        PageAbstract._quick_setup(self, kwargs)

    def toxml(self):
        """
        Returns master page object as LXML Element

        :returns: xml
        :rtype: lxml._Element
        """
        return PageAbstract.toxml(self, True)

    def fromxml(self, xml):
        """
        Set master page attributes according to LXML Element

        :param xml: Master page source as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.fromxml(self, xml, True)

    def fromdefault(self, default=False):
        """
        Set master page attributes according to known defaults.

        :param default: Name of the set of defaults ("a4", "letter")
        :type default: str
        """
        PageAbstract.fromdefault(self, True, default)

    def set_orientation(self, orientation):
        """
        Set master page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: str
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.set_orientation(self, orientation)


class PageSet(PyScribusElement):
    """
    Page set object.

    DOCUMENT/PageSets/Set
    """

    DEFAULTS = {
        "Single Page": {"first": 0, "rows": 1, "columns": 1, "names": []},
        "Facing Pages": {
            "first": 1, "rows": 1, "columns": 2,
            "names": ["Left Page", "Right Page"]
        },
        "3-Fold": {
            "first": 0, "rows": 1, "columns": 3,
            "names": ["Left Page", "Middle", "Right Page"]
        },
        "4-Fold": {
            "first": 0, "rows": 1, "columns": 4,
            "names": [
                "Left Page", "Middle Left", "Middle Right", "Right Page"
            ]
        }
    }

    def __init__(self):
        PyScribusElement.__init__(self)

        self.pyscribus_defaults = [k for k in PageSet.DEFAULTS.keys()]

        self.name = ""
        self.first_page = 0
        self.rows = 0
        self.columns = 0
        self.pages = []

    def fromdefault(self, default):
        """
        """

        if default in self.pyscribus_defaults:

            accurate = PageSet.DEFAULTS[default]

            self.name = default
            self.rows = accurate["rows"]
            self.pages = accurate["names"]
            self.columns = accurate["columns"]
            self.first_page = accurate["first"]

            return True
        else:
            return False

    def toxml(self):
        """
        :returns: xml
        :rtype: lxml._Element
        """

        xml = ET.Element("Set")

        #------------------------------------------------------

        xml.attrib["Name"] = self.name
        xml.attrib["FirstPage"] = str(self.first_page)
        xml.attrib["Rows"] = str(self.rows)
        xml.attrib["Columns"] = str(self.columns)

        #--- Page names ---------------------------------------

        if self.pages:

            for n in self.pages:

                pnx = ET.Element("PageNames")
                pnx.attrib["Name"] = n
                xml.append(pnx)

        #--- FIXME This exports undocumented attributes -------

        try:
            xml = undocumented_to_xml(xml, self.undocumented, no_none=True)
        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromxml(self, xml):
        """
        :param xml: Page set as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """

        if xml.tag == "Set":

            #------------------------------------------------------

            name = xml.get("Name")
            if name is not None:
                self.name = name

            #------------------------------------------------------

            for att in ["FirstPage", "Rows", "Columns"]:

                atx = xml.get(att)
                if atx is not None:

                    try:
                        atx = int(atx)
                    except ValueError:
                        raise exception.InsaneSLAValue(
                            "Page set @{} must be a number.".format(
                                atx
                            )
                        )

            first = xml.get("FirstPage")
            if first is not None:
                self.first_page = int(first)

            rows = xml.get("Rows")
            if rows is not None:
                self.rows = int(rows)

            columns = xml.get("Columns")
            if columns is not None:
                self.columns = int(columns)

            #--- Page names ---------------------------------------

            for child in xml:
                if child.tag == "PageNames":
                    n = child.get("Name")

                    if n not in self.pages:
                        self.pages.append(n)

            #--- FIXME This records undocumented attributes -------

            self.undocumented = undocumented_to_python(
                xml, ["GapBelow", "GapHorizontal", "GapVertical"]
            )

            return True
        else:
            return False

# vim:set shiftwidth=4 softtabstop=4 spl=en:
